# FILE: bug76.tcl
#
#	Patches for Tcl 7.6
#
# Last modified on: $Date: 1998/06/30 19:25:38 $
# Last modified by: $Author: dgp $
#
# This file contains Tcl code which when sourced in a Tcl 7.6 interpreter
# will patch bugs in Tcl 7.6 fixed in later versions of Tcl.  It is not
# meant to be complete.  Bugs are patched here only as we discover the need
# to do so in writing other OOMMF software.

Oc_Log Log "Patching Tcl 7.6 bugs..." status

# There is a bug in [clock format] (fixed in Tcl 7.6p2) which causes it
# to fail if the value of the -format option does not contain a '%'.
rename clock Tcl7.6_clock
;proc clock {option args} {
    switch -exact $option {
        f      -
        fo     -
        for    -
        form   -
        forma  -
        format {
            set clockValue [lindex $args 0]
            array set tmp [lrange $args 1 end]
            if {[catch {regexp % $tmp(-format)} hasPercent] || $hasPercent} {
                uplevel Tcl7.6_clock $option $args
            } else {
                return $tmp(-format)
            }
        }
        default {
            uplevel Tcl7.6_clock $option $args
        }
    }
}

# The subcommands nativename and attributes were added to
# the Tcl command 'file' in Tcl version 8.0.  Here is an approximation
# for earlier Tcl versions:
rename file Tcl7.6_file
;proc file {option args} {
    # Not really the right way to accept unique abbreviations
    switch -glob $option {
        att        -
        attr       -
        attri      -
        attrib     -
        attribu    -
        attribut   -
        attribute  -
        attributes {
            return -code error \
                    "Tcl [info tclversion] does not support \[file attributes\].
	Upgrade to Tcl 8.0 to use it."
        }
        n          -
        na         -
        nat        -
        nati       -
        nativ      -
        native     -
        nativen    -
        nativena   -
        nativenam  -
        nativename {
            if {![llength $args]} {
                return -code error "wrong # args: should be\
                        \"file nativename name ?arg ...?\""
            }
            set fcomps [file split [lindex $args 0]]
            # Take care of tilde substitution
            set first [lindex $fcomps 0]
            if {[string match ~* $first]} {
                set first [file join [file dirname $first] [file tail $first]]
            }
            set result [eval file join [list $first] [lrange $fcomps 1 end]]
            global tcl_platform
            if {[string match windows $tcl_platform(platform)]} {
                regsub -all -- / $result \\ result
            }
            return $result
        }
        default {
            uplevel Tcl7.6_file $option $args
        }
    }
}

# The 'package unknown' handler distributed with Tcl 8.0 is far superior
# to those of earlier Tcl versions.  It is able to withstand errors in
# pkgIndex.tcl files.  It also uses the search path introduced in Tcl 7.6.
#
# The proc 'Oc_PkgUnknown' below is derived from the proc 
# 'tclPkgUnknown' distributed with Tcl 8.0.  
# NOMAC -- It has had Macintosh platform support stripped out.
#
# If the local Tcl/Tk installation uses a customized 'package unknown'
# command, then the Oc extension will also have to be customized to
# use that command.
;proc Oc_PkgUnknown {name version {exact {}}} {
    global auto_path tcl_platform env

    if ![info exists auto_path] {
	return
    }
    for {set i [expr [llength $auto_path] - 1]} {$i >= 0} {incr i -1} {
	# we can't use glob in safe interps, so enclose the following
	# in a catch statement
	catch {
	    foreach file [glob -nocomplain [file join [lindex $auto_path $i] \
		    * pkgIndex.tcl]] {
		set dir [file dirname $file]
		if [catch {source $file} msg] {
		    tclLog "error reading package index file $file: $msg"
		}
	    }
        }
	set dir [lindex $auto_path $i]
	set file [file join $dir pkgIndex.tcl]
	# safe interps usually don't have "file readable", nor stderr channel
	if {[interp issafe] || [file readable $file]} {
	    if {[catch {source $file} msg] && ![interp issafe]}  {
		tclLog "error reading package index file $file: $msg"
	    }
	}
    }
}
package unknown Oc_PkgUnknown

# If Tk is loaded...
# Patch a bug MJD found in the Tk 4.2 distribution
if { [info exists tk_version] && ($tk_version == 4.2) } {
    # The distribution tkMenuFind source seems to be broken: it has
    # comments in an illegal position inside the 'switch' statement.
    # The following code is from Tk 4.2, but with the comments moved.
    proc tkMenuFind { w char } {
	global tkPriv
	set char [string tolower $char]
	
	foreach child [winfo child $w] {
	    switch [winfo class $child] {
		Menubutton {
		    set char2 [string index [$child cget -text]  \
			    [$child cget -underline]]
		    if {([string compare $char \
			    [string tolower $char2]] == 0) \
			    || ($char == "")} {
			if {[$child cget -state] != "disabled"} {
			    return $child
			}
		    }
		}
		default {
		    # The tag above used to be "Frame", but it was changed so
		    # that the code would work with Itcl 2.0, which apparently
		    # uses other classes of widgets to hold menubuttons.
		    set match [tkMenuFind $child $char]
		    if {$match != ""} {
			return $match
		    }
		}
	    }
	}
	return {}
    }
}

# On pre-Tcl 8.0 + Windows, use "fixed" font by default.
# (Not sure we want to keep this...)
if {[llength [info commands option]] && \
	[string match windows $tcl_platform(platform)]} {
    catch { option add *font "fixed" }
}

# Continue with bug patches for the next Tcl version
source [file join [file dirname [info script]] bug80.tcl]

